#Final Project for CHE210D
#BRAD SPATOLA
#6/10/09

#import python modules
import numpy as np
import time

#import compiled fortran library
import fplib

import sys
import os

import atomwrite


#whether or not to use the 3d visualization
UseVisual = False
if UseVisual:
    #import custom visualization library
    import atomvis


#========== GLOBAL VARIABLES ==========
#distance cutoff for pairwise interactions
Cut = 2.5

#NOTE:
#everything below assumes unit atomic masses,
#such that forces = accelerations.

import glob
#make a list of all pdb file in the working path using the glob module
a = glob.glob("*.pdb")


def RadiusofGyration(Pos, N):
    """Returns a float with the corresponding radius of gyration
    Input:  Pos
    Output:  Rg"""
    
    Rg = float
    #N = len(Pos)
    r = []

    rbar = Pos.mean(axis=0) 
    Rgsquared = 1./N*(np.sum((Pos-rbar)**2.))
    Rg = np.sqrt(Rgsquared)

    return Rg

def CalcEnergy(Pos):
    """Wrapper for energy calculation.
Input:
    Pos: (N,3) position array
Output:
    PEnergy: potential energy, float
"""
    global M, Cut
    return fplib.calcenergy(Pos, Cut)


def CalcEnergyForces(Pos, Forces):
    """Wrapper for energy and force calculation.
Input:
    Pos: (N,3) position array
    Forces: (N,3) force array
Output:
    PEnergy: potential energy, float
    Forces: (N,3) force array (updated)
"""
    global Cut
    return fplib.calcenergyforces(Pos, Cut, Forces)


def LineSearch(Pos, Dir, dx, EFracTol,
               Accel = 1.5, MaxInc = 10., MaxIter = 10000):
    """Performs a line search along direction Dir.
Input:
    Pos: starting positions, (N,3) array
    Dir: (N,3) array of gradient direction
    dx: initial step amount
    EFracTol: fractional energy tolerance
    Accel: acceleration factor
    MaxInc: the maximum increase in energy for bracketing
    MaxIter: maximum number of iteration steps
Output:
    PEnergy: value of potential energy at minimum along Dir
    Pos: minimum energy (N,3) position array along Dir
"""
    #start the iteration counter
    Iter = 0

    #find the normalized direction    
    NormDir = Dir / np.sqrt(np.sum(Dir * Dir))

    #take the first two steps and compute energies    
    Dists = [0., dx]
    PEs = [CalcEnergy(Pos + NormDir * x) for x in Dists]
    
    #if the second point is not downhill in energy, back
    #off and take a shorter step until we find one
    while PEs[1] > PEs[0]:
        Iter += 1
        dx = dx * 0.5
        Dists[1] = dx
        PEs[1] = CalcEnergy(Pos + NormDir * dx)
        
    #find a third point
    Dists = Dists + [2. * dx]
    PEs = PEs + [CalcEnergy(Pos + NormDir * 2. * dx)]
    
    #keep stepping forward until the third point is higher
    #in energy; then we have bracketed a minimum
    while PEs[2] < PEs[1]:
        Iter += 1
            
        #find a fourth point and evaluate energy
        Dists = Dists + [Dists[-1] + dx]
        PEs = PEs + [CalcEnergy(Pos + NormDir * Dists[-1])]

        #check if we increased too much in energy; if so, back off
        if (PEs[3] - PEs[0]) > MaxInc * (PEs[0] - PEs[2]):
            PEs = PEs[:3]
            Dists = Dists[:3]
            dx = dx * 0.5
        else:
            #shift all of the points over
            PEs = PEs[-3:]
            Dists = Dists[-3:]
            dx = dx * Accel
            
    #we've bracketed a minimum; now we want to find it to high
    #accuracy
    OldPE3 = 1.e300
    while True:
        Iter += 1
        if Iter > MaxIter:
            print "Warning: maximum number of iterations reached in line search."
            break
            
        #store distances for ease of code-reading
        d0, d1, d2 = Dists
        PE0, PE1, PE2 = PEs

        #use a parobolic approximation to estimate the location
        #of the minimum
        d10 = d0 - d1
        d12 = d2 - d1
        Num = d12*d12*(PE0-PE1) - d10*d10*(PE2-PE1)
        Dem = d12*(PE0-PE1) - d10*(PE2-PE1)
        if Dem == 0:
            #parabolic extrapolation won't work; set new dist = 0 
            d3 = 0
        else:
            #location of parabolic minimum
            d3 = d1 + 0.5 * Num / Dem
            
        #compute the new potential energy
        PE3 = CalcEnergy(Pos + NormDir * d3)
        
        #sometimes the parabolic approximation can fail;
        #check if d3 is out of range < d0 or > d2 or the new energy is higher
        if d3 < d0 or d3 > d2 or PE3 > PE0 or PE3 > PE1 or PE3 > PE2:
            #instead, just compute the new distance by bisecting two
            #of the existing points along the line search
            if abs(d2 - d1) > abs(d0 - d1):
                d3 = 0.5 * (d2 + d1)
            else:
                d3 = 0.5 * (d0 + d1)
            PE3 = CalcEnergy(Pos + NormDir * d3)
            
        #decide which three points to keep; we want to keep
        #the three that are closest to the minimum
        if d3 < d1:
            if PE3 < PE1:
                #get rid of point 2
                Dists, PEs = [d0, d3, d1], [PE0, PE3, PE1]
            else:
                #get rid of point 0
                Dists, PEs = [d3, d1, d2], [PE3, PE1, PE2]
        else:
            if PE3 < PE1:
                #get rid of point 0
                Dists, PEs = [d1, d3, d2], [PE1, PE3, PE2]
            else:
                #get rid of point 2
                Dists, PEs = [d0, d1, d3], [PE0, PE1, PE3]
                
        #check how much we've changed
        if abs(OldPE3 - PE3) < EFracTol * abs(PE3):
            #the fractional change is less than the tolerance,
            #so we are done and can exit the loop
            break
        OldPE3 = PE3

    #return the position array at the minimum (point 1)        
    PosMin = Pos + NormDir * Dists[1]
    PEMin = PEs[1]

    #if using visualization, update the display
    if UseVisual:
        if atomvis.Initialized:
            #update the positions
            atomvis.Update(PosMin)
        else:
            #initialize the visualization window
            atomvis.Init(PosMin)
        
    return PEMin, PosMin

        
def ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG):
    """Performs a conjugate gradient search.
Input:
    Pos: starting positions, (N,3) array
    dx: initial step amount
    EFracTolLS: fractional energy tolerance for line search
    EFracTolCG: fractional energy tolerance for conjugate gradient
Output:
    PEnergy: value of potential energy at minimum
    Pos: minimum energy (N,3) position array 
"""
    PE, Forces = CalcEnergyForces(Pos, np.zeros_like(Pos))
    Dir = Forces
    OldPE = 1.e300
    while abs(PE - OldPE) > EFracTolCG * abs(PE):
        OldPE = PE
        PE, Pos = LineSearch(Pos, Dir, dx, EFracTolLS)
        OldForces = Forces.copy()
        PE, Forces = CalcEnergyForces(Pos, Forces)
        Gamma = np.sum((Forces - OldForces) * Forces) / np.sum(OldForces * OldForces)
        Dir = Forces + Gamma *  Dir
    return PE, Pos


def InitPositions(N):
    """Returns an array of initial positions of each atom in
    the chain.
Input:
    N: number of atoms
Output:
    Pos: (N,3) array of positions
    """
    #make the position array
    Pos = np.zeros((N,3), float)
    #make a list of numbers for the z positions of the chain
    rz = np.arange(0.75, N, 0.75, dtype=float) + np.random.rand()
    #print r
    #loop through z positions in lattice until done
    #for every atom in the system
    i = 0
    for z in rz:
            Pos[i] = np.array([3.*np.random.rand()-3.*np.random.rand(), 3.*np.random.rand()-3.*np.random.rand(),z], float)
            i += 1
            #if done placing atoms, return
            if i >= N:
                return Pos
    return Pos

def RescaleVelocities(Vel, T):
    """Rescales velocities in the system to the target temperature.
Input:
    Vel: (N,3) array of atomic velocities
    T: target temperature
Output:
    Vel: same as above
"""
    #recenter to zero net momentum (assuming all masses same)
    Vel = Vel - Vel.mean(axis=0)
    #find the total kinetic energy
    KE = 0.5 * np.sum(Vel * Vel)
    #find velocity scale factor from ratios of kinetic energy
    #VScale = np.sqrt(1.5 * len(Vel) * T / KE)
    #Advanced technique: Berendsen Thermostat
    tau = 0.01
    VScale = np.sqrt(1. + dt/tau*((1.5 * len(Vel) * T / KE)-1.))
    Vel = Vel * VScale
    return Vel  


def InitVelocities(N, T):
    """Returns an initial random velocity set.
Input:
    N: number of atoms
    T: target temperature
Output:
    Vel: (N,3) array of atomic velocities
"""
    Vel = np.random.rand(N, 3)
    Vel = RescaleVelocities(Vel, T)
    return Vel


def InitAccel(Pos):
    """Returns the initial acceleration array.
Input:
    Pos: (N,3) array of atomic positions
Output:
    Accel: (N,3) array of acceleration vectors
"""
    #get the acceleration from the forces
    PEnergy, Accel = CalcEnergyForces(Pos, np.zeros_like(Pos))
    return Accel


def InstTemp(Vel):
    """Returns the instantaneous temperature.
Input:
    Vel: (N,3) array of atomic velocities
Output:
    Tinst: float
"""
    return np.sum(Vel * Vel) / (3. * len(Vel))

def EndToEndL(Pos):
    """Returns the end-to-end length of the chain
Input:
    Pos: (N,3) array of atomic positions
Output:
    r: float
"""
    return np.sqrt((Pos[29,2]-Pos[0,2])**2.)


#Define the number of particles, N
N = 30
print N
#No length scaling was used in my simulation
#L = (N/0.8)**(1./3.)
#print L
EFracTolLS = 1.e-8
EFracTolCG = 1.e-10
dx = 0.0001

rc = 2.5
dt = 0.0001

time = 0.
T = 1.0

#print T
Pos = InitPositions(N)
print Pos
Names = []
x = 0
#Call the atoms in the chain O so that they are easier to select in Chimera
while x<N:
    Names.append('O')
    x += 1
    
Pdb = atomwrite.pdbfile("fptest.pdb", AtomNames = Names)

Vel = InitVelocities(N, T)
#print Vel
Accel = InitAccel(Pos)
#print Accel
PEnergy = []

#Histogram code HERE
rmax = EndToEndL(Pos)
rmin = 0.
#Calculate the bin width with m bins
m = 25.
dr = (rmax-rmin)/m


#Energy minimize the initial configuration using the Conjugate Gradient Method
#This allows for greater numeric stability when the MD integration starts
(PE, Pos) = ConjugateGradient(Pos, dx, EFracTolLS, EFracTolCG)
print "CONJ GRAD DONE"
print PE
print Pos

i = 0
j = 0
TEnergy = []
RadiusGy = []
#NStepsEquil1

while time < 1.000:
    if i % 1 == 0:
        #The velocity is rescaled every step using the Berendsen Thermostat
        Vel = RescaleVelocities(Vel, T)
        i = 0
    if j % 1000 == 0:
        #Write the Positions to a PDB file for viewing in Chimera
        Pdb.write(Pos)
        j = 0
    #Call the Velocity Verlet algorithm from Fortran         
    (Pos, Vel, Accel, KEnergy, PEnergy) = fplib.vvintegrate(Pos, Vel, Accel, rc, dt)
    time += 0.0001
    i += 1
    j += 1

print Pos
print PEnergy
#print Vel


i = 0
j = 0
while time < 10.000:
    if i % 1 == 0:
        Vel = RescaleVelocities(Vel, T)
        #print E, PEnergy
        i = 0
    if j % 1000 == 0:
        #Write the Positions to a PDB file for viewing in Chimera
        Pdb.write(Pos)
        j = 0
    #Call the Velocity Verlet algorithm from Fortran
    (Pos, Vel, Accel, KEnergy, PEnergy) = fplib.vvintegrate(Pos, Vel, Accel, rc, dt)
    E = KEnergy+PEnergy
    TEnergy.append(E)
    
    time += 0.0001
    i += 1
    j += 1



AvgTE = np.mean(TEnergy)
print "%.5f  %.5f" % (AvgTE, KEnergy+PEnergy)
Vel = np.sqrt((AvgTE-PEnergy)/KEnergy) * Vel
#just check if the rescaling worked in the next two lines below
(Pos, Vel, Accel, KEnergy, PEnergy) = fplib.vvintegrate(Pos, Vel, Accel, rc, dt)
print "%.5f  %.5f" % (AvgTE, KEnergy+PEnergy)

#NStepsProd
i = 0
j = 0
k = 0
l = 0
R = np.zeros((m-1))
RadiusGy = []
t = []

while time < 100.000:
    (Pos, Vel, Accel, KEnergy, PEnergy) = fplib.vvintegrate(Pos, Vel, Accel, rc, dt)
    E = KEnergy+PEnergy
    TEnergy.append(E)

    if j % 1000 == 0:
        #Calculate the Radius of Gyration
        Rg = RadiusofGyration(Pos, N)
        #Write the Positions to a PDB file for viewing in Chimera
        Pdb.write(Pos)
        RadiusGy.append(Rg)
        print "%.4f     %.8f    %.8f    %.8f" %(time, PEnergy, E, Rg)
        
        #print RadiusGy
        #InstT = InstTemp(Vel)
        #print InstT, T
        j = 0
    
    #Histogram code here
    if l % 100 == 0:
        r = EndToEndL(Pos)
        #print r
        while k < m-1:        
            if r > dr*k and r < (dr*(k+1)):
                R[k] = R[k] + 1
                k = m
            else:
                k += 1
        l = 0
        k = 0
        
    time += 0.0001
    j += 1
    l += 1
print R
print "Mean Radius of Gyration at %f" %(T)
print np.mean(RadiusGy)

Pdb.close()

